/*
 * on Windows you can build this with "cl /Oi /O2 rust.cc"
 * on GCC platforms use "g++ -O2 -march=native rust.cc"
 */

#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>
#include <emmintrin.h>
#include <immintrin.h>

typedef uint8_t u8;
typedef uint16_t u16;
typedef uint64_t u64;

#if defined( _WIN32 )

#include <windows.h>

inline double get_time() {
        LARGE_INTEGER counter;
        QueryPerformanceCounter( &counter );

        LARGE_INTEGER freq;
        QueryPerformanceFrequency( &freq );

        return double( counter.QuadPart ) / double( freq.QuadPart );
}

#elif defined( __APPLE__ )

#include <mach/mach_time.h>

inline double get_time() {
        mach_timebase_info_data_t freq;
        mach_timebase_info( &freq );

        u64 ticks = mach_absolute_time();

        return double( ticks ) * ( freq.numer / 1e9 ) * double( freq.denom );
}

#else

#include <time.h>

inline double get_time() {
        struct timespec ts;
        clock_gettime( CLOCK_MONOTONIC, &ts );

        return double( ts.tv_sec ) + ts.tv_nsec / 1e9;
}

#endif

#if defined( _MSC_VER )
#define popcnt64 __popcnt64
#else
#define popcnt64 __builtin_popcountll 
#endif

size_t count_slow( const u8 * haystack, size_t n, u8 needle ) {
	size_t c = 0;
	for( size_t i = 0; i < n; i++ ) {
		if( haystack[ i ] == needle ) {
			c++;
		}
	}
	return c;
}

size_t count_fast( const u8 * haystack, size_t n, u8 needle ) {
	const u8 * one_past_end = haystack + n;
	size_t c = 0;

	// peel
	while( uintptr_t( haystack ) % 16 != 0 && haystack < one_past_end ) {
		if( *haystack == needle ) {
			c++;
		}
		haystack++;
	}

	__m128i needles = _mm_set1_epi8( needle );
	while( haystack < one_past_end - 16 ) {
		__m128i counts = _mm_setzero_si128();

		for( int i = 0; i < 255; i++ ) {
			if( haystack >= one_past_end - 16 ) {
				break;
			}
			__m128i chunk = _mm_load_si128( ( const __m128i * ) haystack );
			__m128i cmp = _mm_cmpeq_epi8( needles, chunk );
			counts = _mm_sub_epi8( counts, cmp );
			haystack += 16;
		}

		__m128i sums = _mm_sad_epu8( counts, _mm_setzero_si128() );
		u16 sums_[ 8 ];
		_mm_storeu_si128( ( __m128i * ) sums_, sums );
		c += sums_[ 0 ] + sums_[ 4 ];
	}

	// remainder
	while( haystack < one_past_end ) {
		if( *haystack == needle ) {
			c++;
		}
		haystack++;
	}

	return c;
}

size_t count_fast_unrolled( const u8 * haystack, size_t n, u8 needle ) {
	const u8 * one_past_end = haystack + n;
	size_t c = 0;

	// peel
	while( uintptr_t( haystack ) % 16 != 0 && haystack < one_past_end ) {
		if( *haystack == needle ) {
			c++;
		}
		haystack++;
	}

	__m128i needles = _mm_set1_epi8( needle );
	while( haystack < one_past_end - 64 ) {
		__m128i counts0 = _mm_setzero_si128();
		__m128i counts1 = _mm_setzero_si128();
		__m128i counts2 = _mm_setzero_si128();
		__m128i counts3 = _mm_setzero_si128();

		for( int i = 0; i < 255; i++ ) {
			if( haystack >= one_past_end - 64 ) {
				break;
			}

			{
				__m128i chunk = _mm_load_si128( ( const __m128i * ) haystack );
				__m128i cmp = _mm_cmpeq_epi8( needles, chunk );
				counts0 = _mm_sub_epi8( counts0, cmp );
				haystack += 16;
			}

			{
				__m128i chunk = _mm_load_si128( ( const __m128i * ) haystack );
				__m128i cmp = _mm_cmpeq_epi8( needles, chunk );
				counts1 = _mm_sub_epi8( counts1, cmp );
				haystack += 16;
			}

			{
				__m128i chunk = _mm_load_si128( ( const __m128i * ) haystack );
				__m128i cmp = _mm_cmpeq_epi8( needles, chunk );
				counts2 = _mm_sub_epi8( counts2, cmp );
				haystack += 16;
			}

			{
				__m128i chunk = _mm_load_si128( ( const __m128i * ) haystack );
				__m128i cmp = _mm_cmpeq_epi8( needles, chunk );
				counts3 = _mm_sub_epi8( counts3, cmp );
				haystack += 16;
			}
		}

		__m128i zeroes = _mm_setzero_si128();
		__m128i sums0 = _mm_sad_epu8( counts0, zeroes );
		__m128i sums1 = _mm_sad_epu8( counts1, zeroes );
		__m128i sums2 = _mm_sad_epu8( counts2, zeroes );
		__m128i sums3 = _mm_sad_epu8( counts3, zeroes );
		u16 sums0_[ 8 ];
		u16 sums1_[ 8 ];
		u16 sums2_[ 8 ];
		u16 sums3_[ 8 ];
		_mm_storeu_si128( ( __m128i * ) sums0_, sums0 );
		_mm_storeu_si128( ( __m128i * ) sums1_, sums1 );
		_mm_storeu_si128( ( __m128i * ) sums2_, sums2 );
		_mm_storeu_si128( ( __m128i * ) sums3_, sums3 );
		c +=
			sums0_[ 0 ] + sums0_[ 4 ] +
			sums1_[ 0 ] + sums1_[ 4 ] +
			sums2_[ 0 ] + sums2_[ 4 ] +
			sums3_[ 0 ] + sums3_[ 4 ];
	}

	// remainder
	while( haystack < one_past_end ) {
		if( *haystack == needle ) {
			c++;
		}
		haystack++;
	}

	return c;
}

int main() {
	// build an array of 10000 Lorem Ipsums
	int N = 100000;
	u8 lorem[] = "Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt ut labore et dolore magna aliqua. Ut enim ad minim veniam, quis nostrud exercitation ullamco laboris nisi ut aliquip ex ea commodo consequat. Duis aute irure dolor in reprehenderit in voluptate velit efast cillum dolore eu fugiat nulla pariatur. Excepteur sint occaecat cupidatat non proident, sunt in culpa qui officia deserunt mollit anim id est laborum.";

	size_t sz = sizeof( lorem ) * N;
	u8 * lorems = ( u8 * ) malloc( sz );
	if( lorems == NULL ) return 0;

	for( int i = 0; i < N; i++ ) {
		memcpy( lorems + i * sizeof( lorem ), lorem, sizeof( lorem ) );
	}

	for( int i = 0; i < 10; i++ ) {
		double before = get_time();
		size_t c = count_slow( lorems, sz, 'o' );
		double dt = get_time() - before;
		printf( "count_slow found %llu 'o's in %.3fms\n", c, dt * 1000.0 );
	}
	printf( "\n" );

	for( int i = 0; i < 10; i++ ) {
		double before = get_time();
		size_t c = count_fast( lorems, sz, 'o' );
		double dt = get_time() - before;
		printf( "count_fast found %llu 'o's in %.3fms\n", c, dt * 1000.0 );
	}
	printf( "\n" );

	for( int i = 0; i < 10; i++ ) {
		double before = get_time();
		size_t c = count_fast_unrolled( lorems, sz, 'o' );
		double dt = get_time() - before;
		printf( "count_fast_unrolled found %llu 'o's in %.3fms\n", c, dt * 1000.0 );
	}
	printf( "\n" );

	return 0;
}
